// =============================================================================
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
//
// Copyright (C) 2021 - Stéphane Mottelet
//
// This file must be used under the terms of GPL License
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available atsol
// https://www.gnu.org/licenses/gpl-3.0.txt

// basic example
function out = res(t,y,yp)
    out = y+yp;
end
[t,y] = ida(res,[0 5],1,-1);


// SIR MODEL
function res = sir(t,y,yp,bet,gam,N)
    res=[yp(1)+bet/N*y(1)*y(2)
         yp(2)-bet/N*y(1)*y(2)+gam*y(2)  
         y(1)+y(2)+y(3)-N];
end
function [jacy,jacyp] = sir_jac(t,y,yp,bet,gam,N)
    jacy = [ bet/N*y(2)  bet/N*y(1)     0
            -bet/N*y(2) -bet/N*y(1)+gam 0
             1           1              1];
    jacyp = diag([1 1 0]);
end
function out = sir_ev(t,y,yp,bet,gam,N)
    out = yp(2);
end
N=60e6;
gam=1/40;
bet=0.2;
y0=[N-1;1;0];
yp0 = [-bet/N*y0(1)*y0(2);+bet/N*y0(1)*y0(2)-gam*y0(2);gam*y0(2)];
[t,y,yp,te,ye,ype] = ida(list(sir,bet,gam,N),[0 400],y0,yp0,events=list(sir_ev,bet,gam,N));
assert_checkalmostequal(te,114.56018)
assert_checkalmostequal(ype(2),0,0,1e-6)


//with jacobian and extending solution
sol = ida(list(sir,bet,gam,N),[0 200],y0,yp0);
sol2 = ida(list(sir,bet,gam,N),[0 200],y0,yp0,jacobian=list(sir_jac,bet,gam,N));

assert_checkalmostequal(sol(sol.t),sol.y)
sol2 = ida(sol,400);
assert_checkalmostequal(sol2(sol.t),sol.y)
assert_checkalmostequal(sol2(sol2.t),sol2.y)

// compute initial condition
[t,y,yp] = ida(list(sir,bet,gam,N),[0,400],y0,yp0,rtol=1e-14);
y0=[N-1;1;0];
yp0 = [0;0;0];
[tt,yy,yyp]= ida(list(sir,bet,gam,N),[0 400],y0,yp0,rtol=1e-14,calcIc="y0yp0",yIsAlgebric=3);
assert_checkalmostequal(yy(:,$),y(:,$),1e-6);

// Robertson problem
function res = robertsidae(t,y,yp)
    res = [yp(1)+0.04*y(1)-1e4*y(2)*y(3);
    yp(2)-0.04*y(1)+1e4*y(2)*y(3)+3e7*y(2)*y(2);
    y(1)+y(2)+y(3)-1];
end

y0 = [1-1e-3; 0; 1e-3];
yp0 = [-0.0400; 0.0400; 0];
tspan = [0 4e6];
[t,y,yp] = ida(robertsidae,tspan,y0,yp0,h0=1e-6);
solr = ida(robertsidae,tspan,y0,yp0,h0=1e-6);
// compute initial condition of algebric state
y0 = [1-1e-3; 0; 0];
yp0 = [-0.0400; 0.0400; 0];
[t,y,yp] = ida(robertsidae,tspan,y0,yp0,h0=1e-6,calcIc="y0yp0",yIsAlgebric=3);

// Linear DAE
function out = res(t,y,yp,A,B)
    out = A*yp-B*y;
endfunction
A = [2 1; 1 1];
B = [0 1;-1 -1];
y0 = [1;0];
yp0 = A\(B*y0);
sol = ida(list(res,A,B),[0 10],y0,yp0);
assert_checkequal(sol.stats.nfevalsFD,40);
solj = ida(list(res,A,B),[0 10],y0,yp0,jacobian={-B,A});
assert_checkequal(solj.stats.nfevalsFD,0);
[t,y,yp] = ida(list(res,A,B),[0 10],y0,zeros(2,1),calcIc="y0yp0");
assert_checkalmostequal(yp(:,1),yp0);

// fully implicit
function res = weissinger(t,y,yp)
    res=t*y^2 * yp^3 - y^3 * yp^2 + t*(t^2 + 1)*yp - t^2 * y;
end
t0=1;
y0=sqrt(3/2);
yp0=0;
[t,y,yp]=ida(weissinger,[1 10],y0,yp0,calcIc="y0yp0");
assert_checkalmostequal(yp(:,1),0.816496582)

// Dynamic link
// DYNAMIC LINK Residual
rescode=["void chemres(int *n, double *t, double y[], double yd[], double r[])"
         " {"
         "   r[0] = yd[0]+0.04*y[0]-1.0e4*y[1]*y[2];"
         "   r[1] = yd[1]-0.04*y[0]+1.0e4*y[1]*y[2]+3.0e7*y[1]*y[1];"
         "   r[2] = y[0]+y[1]+y[2]-1;"
         " }"];
jaccode=["void chemjac(int *n, double *t, double *cj, double y[], double yd[], double r[], double j[])"
         " {"
         "   /* first column*/"
         "   j[0] = 0.04+*cj;"
         "   j[1] =  -0.04;"
         "   j[2] =  1.0;"
         "    /* second column*/"
         "   j[3] =  -1.0e4*y[2];"
         "   j[4] = +1.0e4*y[2]+2*3.0e7*y[1]+*cj;"
         "   j[5] =  1.0;"
         "    /* third column*/"
         "   j[6] =  -1.0e4*y[1];"
         "   j[7] = +1.0e4*y[1];"
         "   j[8] =  1.0;"
         " }"];
         eventcode=["void chemevent(int *n,double *t,double *y, double *yd, int *ng, double *g, int *term, int *dir)"
             "{ "
             "if (*ng == 0) {"
             "    *ng = 1; "
             "} else if (g == 0) {"
             "    term[0] = 0; dir[0] = 0;"
             "} else {"
             "    g[0] = yd[1];"
             "}}"];         
        cbcode=["int chemcb(int *n,double *t,double *y, double *yd, int *flag)"
                 "{ "
                 "sciprint(""t=%f, y2=%e, yp2=%e\n"",*t,y[1],yd[1]);"
                 "return 0;"
                 "}"];         

y0 = [1-1e-3; 0; 1e-3];
yp0 = [-0.0400; 0.0400; 0];
tspan = [0 4e6];

mputl(["#include ""sciprint.h""";rescode;jaccode;eventcode;cbcode], "mycode.c") //create the C file
ilib_for_link(["chemres","chemjac","chemevent","chemcb"],"mycode.c",[],"c");//compile
exec("loader.sce") //incremental linking

tic;
for i=1:10
[t1,y1,yp1] = ida("chemres",tspan,y0,yp0);
end
et1=toc();
tic;
for i=1:10
[t2,y2,yp2] = ida(robertsidae,tspan,y0,yp0);
end
et2=toc();
tic;
for i=1:10
[t3,y3,yp3] = ida("chemres",tspan,y0,yp0,jacobian="chemjac");
end
et3=toc();

assert_checkequal(t1,t2)
assert_checkequal(y1,y2)
assert_checkequal(yp1,yp2)
// using a C gateway outperforms Scilab macro 
assert_checktrue(et2/et1 > 50);
// true Jacobian really helps the solver
assert_checktrue(et1/et3 > 4);
assert_checktrue(size(t1)/size(t3) > 4)

//test dynamic event
[t3,y3,yp3,te,ye] = ida("chemres",tspan,y0,yp0,jacobian="chemjac",events="chemevent");
assert_checkalmostequal(te,0.0048495040982)
//tes dynamic callback
[t1,y1,yp1] = ida("chemres",tspan,y0,yp0,intcb="chemcb");


